/*
 * Photo Image Print System
 * Copyright (C) 2000-2005 EPSON AVASYS Corporation.
 * Copyright (C) SEIKO EPSON CORPORATION 2000-2005.
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * As a special exception, EPSON AVASYS Corporation gives permission to
 * link the code of this program with libraries which are covered by
 * the EPSON AVASYS Public License and distribute their linked
 * combinations.  You must obey the GNU General Public License in all
 * respects for all of the code used other than the libraries which
 * are covered by EPSON AVASYS Public License.
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "pipscap-data.h"
#include "pipscap-util.h"
#include "pipscap.h"

#define PIPS_PRINTER_PATH "/usr/local/EPAva/printer/"
#define PIPS_RC_FILE "/etc/pipsrc"
#define ENTRY_MAX 48
#define OPTION_MAX 30
#define MODE_OPTION_MAX OPTION_MAX
#define ARRAY_MAX 64
#define SUB_MAX 3

/******************* Define internal function*********************************/
static int priority_table[MODE_OPTION_MAX] = {
	P_INK, P_INKSET, P_PAPER_SOURCE, P_MEDIA_TYPE,
	P_RESOLUTION, P_QUALITY, P_MICRO_WEAVE
        };
static int PrintModeCheck (CapContextList, int, char *);
static int CombineContextList (CapContextList*, CapContextList);
static int CheckOption(ElementList list, int capid, char sub[SUB_MAX][ARRAY_MAX]);
static int IsExist (char *, char **);
static void ReferModeRange (PrintModeList, int, char *, int [], int *);
static int GetDefPrintMode (CapInforPtr, CapContextList *);
/*****************************************************************************/
static int GetDefPrintMode (CapInforPtr cap_infor, CapContextList *context_list)
{
    CapList cap_list = cap_infor->data_list->common_list;
    ElementList element_list;
    PaperSizeList papersize_list = cap_infor->data_list->papersize_list;
    PrintModeList mode_list = cap_infor->data_list->mode_color_list;
    CapContextList mode_context_list, def_list = cap_infor->data_list->name_def_list;

    /* Get default value of common option firstly */
    while (cap_list != NULL)
    {
        element_list = cap_list->list;
        while (element_list != NULL)
        {
            if (!strcmp (element_list->option, "def"))
            {
                if (AddContextListCap (context_list, cap_list->capid, element_list->value))
                    return MEMORY_ERROR;
                break;
            }
            else
                element_list = element_list->next;
        }

        cap_list = cap_list->next;
    }

    /* Set default inkset. */
    if (SearchContextListCap (*context_list, P_INKSET) == NULL) {
	    char *option = NULL;
	    /* GetCapOption (cap_infor, P_INK_CHANGE_SYSTEM, option); */
	    option = SearchContextListCap (cap_infor->data_list->other_list, P_INK_CHANGE_SYSTEM);
	    if (option != NULL && atoi(option) != 0) {
		    cap_list = SearchCapListCap (cap_infor->data_list->cap_list, P_INKSET);
		    if (cap_list != NULL && cap_list->list->option != NULL) {
			    if (AddContextListCap (context_list, P_INKSET, cap_list->list->option))
				    return MEMORY_ERROR;
		    }
	    }
    }
    
    /* Set default color */
    if (AddContextListCap (context_list, P_INK, "COLOR"))
        return MEMORY_ERROR;
    /* Get default value of PaperSize */
    /* Get the first position in list */
    while (papersize_list->next != NULL)
        papersize_list = papersize_list->next;
    if (AddContextListCap (context_list, P_MEDIA_SIZE, papersize_list->ptid))
        return MEMORY_ERROR;
    /* Set default wideprint */
    if (AddContextListCap (context_list, P_WIDE_PRINT, "ON"))
        return MEMORY_ERROR;
    /* Get default print mode */
    while (def_list != NULL)
    {
        if (def_list->capid == P_DEF_MODE)
        {
            cap_infor->modeid = atoi (def_list->option);
            break;
        }
        def_list = def_list->next;
    }
    while ((mode_list != NULL) && (mode_list->modeid != cap_infor->modeid))
        mode_list = mode_list->next;
    mode_context_list = mode_list->list;
    while (mode_context_list != NULL)
    {
	    if (mode_context_list->capid != P_INKSET) {
		    if (AddContextListCap (context_list, mode_context_list->capid, GetValueCap (mode_context_list->option, 1)))
			    return MEMORY_ERROR;
	    }
	    mode_context_list = mode_context_list->next;
    }

    return NO_ERROR;
}


/*****************************************************************************/
CapResultCode
OpenCap (CapHandle *handle, char *printer, int filter_type, char ret_data[])
{
    CapResultCode ret_code = 0;
    CapInforPtr cap_infor;
    CapContextList context_list;
    CapList cap_list;
    char xmlfile[MAX_SIZE], rcfile[] = PIPS_RC_FILE;
    char PRINTER_PACKAGE[MAX_SIZE] = "pips-";

    /* Initialize */
    cap_infor = (CapInforPtr) calloc (1, sizeof (CapInfor));
    if (!cap_infor)
        return MEMORY_ERROR;

    cap_infor->filter_type = filter_type;
    cap_infor->modeid = 1;
    cap_infor->sub_context_list = NULL;
    cap_infor->context_list = NULL;
    cap_infor->data_list = NULL;

    sprintf (xmlfile, "%s%s%s%s%s", PIPS_PRINTER_PATH, printer, "/", printer,".xml");
    if (access (xmlfile, F_OK | R_OK) == -1)
    {
        strcpy (ret_data, xmlfile);
        return FILE_ACCESS_ERROR;
    }

    PipsCapParserPtr parser_state = (PipsCapParserPtr) malloc (sizeof (PipsCapParser));
    if (!parser_state)
        return MEMORY_ERROR;

    strcat (PRINTER_PACKAGE, printer);
    bindtextdomain (PRINTER_PACKAGE, LOCALEDIR);
    bind_textdomain_codeset (PRINTER_PACKAGE, "UTF-8");
    textdomain (PRINTER_PACKAGE);

    parser_state->filter_type = filter_type;
    if ((ret_code = GetDatabase (xmlfile, parser_state)))
    {
        if (parser_state->ret_data)
            strcpy (ret_data, parser_state->ret_data);
        return ret_code;
    }

    cap_infor->data_list = parser_state->data_list;
    printer = SearchContextListCap (cap_infor->data_list->name_def_list, P_MODEL_NAME);
    
    if ((filter_type == PNG_TYPE) || (filter_type == RAW_TYPE)) {
	    if (GetDefPrintMode(cap_infor, &cap_infor->context_list)) {
		    return MEMORY_ERROR;
	    }
	    
	    if (filter_type == PNG_TYPE) {   /* Load pipsrc */
		    LoadPipsrcFile(printer, rcfile, &cap_infor->context_list);
		    
		    /* Concat data_list->(cap_list & common_list) */
		    cap_list = cap_infor->data_list->cap_list;
		    while (cap_list->next != NULL) {
			    cap_list = cap_list->next;
		    }
		    cap_list->next = cap_infor->data_list->common_list;
	    }
    }
    /* Concat data_list->name_def_list, other_list*/
    context_list = cap_infor->data_list->other_list;
    while (context_list->next != NULL)
        context_list = context_list->next;
    context_list->next = cap_infor->data_list->name_def_list;

    *handle = (CapHandle)cap_infor;

    return  ret_code;
}

/*****************************************************************************/
CapResultCode
CloseCap (CapHandle handle)
{
    /* Free database, context data */
    CapInforPtr cap_infor = (CapInforPtr) handle;
    if (cap_infor != NULL)
    {
	if (cap_infor->context_list != NULL)
		free (cap_infor->context_list);
	if (cap_infor->sub_context_list)
		free (cap_infor->sub_context_list);
	if (cap_infor->data_list)
		free (cap_infor->data_list);

	free (cap_infor);
	cap_infor = NULL;
    }
    if (errno)
        return MEMORY_ERROR;

    return NO_ERROR;
}

/*****************************************************************************/
CapResultCode
GetCapOption (CapHandle handle, CapID id, char *option)
{
    CapInforPtr cap_infor = (CapInforPtr) handle;
    char *opt_temp = NULL;
    /* Search context list first */
    if ((opt_temp = SearchContextListCap (cap_infor->context_list, id)))
        strcpy (option, opt_temp);
    else if ((opt_temp = SearchContextListCap (cap_infor->data_list->other_list, id)))
        strcpy (option, opt_temp);/* Search other list second */
    else
        memset (option, '\0', 1);

    return NO_ERROR;
}

/*****************************************************************************/
CapResultCode
SetCapOption (CapHandle handle, CapID capid, char *option)
{
    CapInforPtr cap_infor = (CapInforPtr) handle;

    if (capid == P_DEF_MODE)
    {
	/* Set default values. */
	return GetDefPrintMode (cap_infor, &cap_infor->sub_context_list);
    }

    if (AddContextListCap (&cap_infor->sub_context_list, capid, option))
        return MEMORY_ERROR;

    return NO_ERROR;
}

/*****************************************************************************/
static int
CheckOption(ElementList list, int capid, char sub[SUB_MAX][ARRAY_MAX])
{
    int i, j, ni, min, max, value;
    ElementList element_list = list;
    ni = 0;
    switch (capid)
    {
    case P_STRENGTH_CYAN:
    case P_STRENGTH_MAGENTA:
    case P_STRENGTH_YELLOW:
        ni ++;
    case P_MARGIN_X:
    case P_MARGIN_Y:
        ni ++;
    case P_SCALE:
    case P_BRIGHTNESS:
    case P_CONTRAST:
    case P_SATURATION:
        ni++;
        for (i = 0; i < ni; i++)
        {
            for (j = 0; sub[i][j] != (char) NULL; j++)
            {
                if ((sub[i][j] < '0') || (sub[i][j] > '9'))
                {
                    if ((sub[i][j] != '-') && (j != 0))
                        return 1;
                }
            }

            value = (int) atoi (sub[i]);
            min = (int) atoi (SearchElementListCap (element_list, "min"));
            if ((capid == P_MARGIN_X) || (capid == P_MARGIN_Y))
            {
                /* Check current paper size correspondence with paper source */
                /* max = GetMaxPaperSize() for left & bottom margin */
                max = 1000; /* Temporary value */
                if ((value < min) || (value > max))
                    return 1;
            }
            else
            {
                max = (int) atoi (SearchElementListCap (element_list, "max"));
                if ((value < min) || (value > max))
                    return 1;
            }
        }
        return 0;
    default:
        break;
    }
    /* Check other feature */
    while (element_list != NULL)
    {
        if (!strcmp (element_list->option, sub[0]))
            return 0;
        element_list = element_list->next;
    }
    /* This feature does not support this option */
    return 1;
}

/*****************************************************************************/
CapResultCode
SetCapList (CapHandle handle, char *argv[], int nargc, char err_data[])
{
    CapInforPtr cap_infor = (CapInforPtr) handle;
    CapList cap_list;
    const char *gopts[] =
        {
            "f", /* Select filter type */
            "p", /* Select printer name */
            "o", /* Output file name */
            "r", /* Temp pipsrc file */
            "at"/* Attribute bitmap input file path */
        };
    char *filename = NULL;
    int i, ni, argc, checked, common_checked = 0, detail_checked = 0;

    /* Get standard parameter & capid */
    for (argc = 1; argc < nargc; argc++)
    {
        if (argv[argc][0] == '-')
        {
            char sub_arg[SUB_MAX][ARRAY_MAX];
            checked = 0;
            strcpy (err_data, argv[argc]);
            cap_list = cap_infor->data_list->cap_list;
            while (cap_list != NULL)
            {
                /* Check option */
                if (!strcmp (argv[argc] + 1, cap_list->param)) /* Parameters only have 2 character */
                {
                    if ((cap_list->capid == P_STRENGTH_CYAN) || (cap_list->capid == P_STRENGTH_MAGENTA) || (cap_list->capid == P_STRENGTH_YELLOW))
                        ni = 3;
                    else if ((cap_list->capid == P_MARGIN_X) || (cap_list->capid == P_MARGIN_Y))
                        ni = 2;
                    else
                        ni = 1;

                    for (i = 0; i < ni; i++)
                    { /* error check */
                        strcat (err_data, " ");
                        if (argc + 1 < nargc)
                            strcat (err_data, argv[argc + 1]);

                        if ((++argc < nargc) && ((argv[argc][0] != '-') || (isdigit (argv[argc][1]))))
                        {
                            if ((cap_infor->filter_type == 2) && (cap_list->capid == P_MEDIA_SIZE))
                            {
                                char *temp = strtok (argv[argc], "_");
                                strcpy (sub_arg[0], temp);
                                temp = strtok (NULL, "_");
                                strcpy (sub_arg[1], temp);
                            }
                            else
                                strcpy (sub_arg[i], argv[argc]);
                        }
                        else
                            return OPTION_LESS_ERROR;
                    }
                    /* Check value */

                    if ((cap_infor->filter_type == 2) && (cap_list->capid == P_MEDIA_SIZE))
                    { /* Special case */
                        if (CheckOption (cap_list->list, cap_list->capid, sub_arg))
                            return OPTION_VALUES_ERROR;

                        CapList temp_cap_list = cap_infor->data_list->cap_list;
                        while ((temp_cap_list != NULL) && (temp_cap_list->capid != P_PAPER_SOURCE))
                            temp_cap_list = temp_cap_list->next;

                        if (!strncmp (sub_arg[1], "EX", 2)) /* For wideprint case */
                            strcpy (sub_arg[2], sub_arg[1] + 2);
                        else
                            strcpy (sub_arg[2], sub_arg[1]);

                        if (CheckOption (temp_cap_list->list, P_PAPER_SOURCE, &sub_arg[2]))
                            return OPTION_VALUES_ERROR;
                    }
                    else if (CheckOption (cap_list->list, cap_list->capid, sub_arg))
                        return OPTION_VALUES_ERROR;

                    /* Set to sub context list */
                    switch (ni)
                    {
                    case 3: /* Density */
                        if (AddContextListCap (&cap_infor->sub_context_list, P_STRENGTH_CYAN, sub_arg[0])
                                ||AddContextListCap (&cap_infor->sub_context_list, P_STRENGTH_MAGENTA, sub_arg[1])
                                ||AddContextListCap (&cap_infor->sub_context_list, P_STRENGTH_YELLOW, sub_arg[2]))
                            return MEMORY_ERROR;
                        break;

                    case 2: /* Margin */
                        if (AddContextListCap (&cap_infor->sub_context_list, P_MARGIN_X, sub_arg[0])
                                ||AddContextListCap (&cap_infor->sub_context_list, P_MARGIN_Y, sub_arg[1]))
                            return MEMORY_ERROR;
                        break;

                    case 1: /* Others */
                        if (AddContextListCap (&cap_infor->sub_context_list, cap_list->capid, sub_arg[0]))
                            return MEMORY_ERROR;
                        if ((cap_infor->filter_type == 2) && (cap_list->capid == P_MEDIA_SIZE))
                        {
                            if (AddContextListCap (&cap_infor->sub_context_list, P_PAPER_SOURCE, sub_arg[2]))
                                return MEMORY_ERROR;
                            char *wide_print = "OFF";
                            if (strcmp (sub_arg[1], sub_arg[2]))
                                wide_print = "ON";
                            if (AddContextListCap (&cap_infor->sub_context_list, P_WIDE_PRINT, wide_print))
                                return MEMORY_ERROR;
                        }
                        break;
                    }

                    checked = 1;
                    if (common_checked)
                        return COMBINE_OPTION_ERROR;
                    detail_checked = 1;

                    break;
                }
                cap_list = cap_list->next;
            }

            /* Check global option */
            if (!checked)
                for (i = 0; i < 5; i++)
                {
                    if (!strcmp (argv[argc] + 1, gopts[i]))
                    {
                        /* error check */
                        if ((++argc < nargc) && (argv[argc][0] != '-'))
                        {
                            strcat (err_data, " ");
                            strcat (err_data, argv[argc]);
                        }
                        else
                            return OPTION_LESS_ERROR;

                        checked = 1;
                        if (!strcmp (gopts[i], "r"))
                        {
                            if (detail_checked)
                                return COMBINE_OPTION_ERROR;
                            common_checked = 1;
                        }
                        else if (!strcmp (gopts[i], "o"))
                        {
                            if (AddContextListCap (&cap_infor->data_list->other_list, P_OUTPUT_FILE, argv[argc]))
                                return MEMORY_ERROR;
                        }
                        else if ((!strcmp (gopts[i], "at")) && (cap_infor->filter_type == 1))
                        {
                            if (AddContextListCap (&cap_infor->data_list->other_list, P_ATBMP_FILE, argv[argc]))
                                return MEMORY_ERROR;
                        }
                        break;
                    }
                }

            if (!checked)
                return INVALID_OPTION_ERROR;
        }
        else if (filename == NULL)
        {
            filename = argv[argc];
            if (AddContextListCap (&cap_infor->data_list->other_list, P_INPUT_FILE, argv[argc]))
                return MEMORY_ERROR;
        }
        else if (strcmp (filename, argv[argc]))
        {
            sprintf (err_data, "\"%s\" \"%s\"", filename, argv[argc]);
            return DUB_FILENAME_ERROR;
        }
    }

    return NO_ERROR;
}

/*****************************************************************************/
static void
ReferModeRange (PrintModeList all_mode_list, int capid, char *option, int flag[], int *passed_mode)
{
    int j = 0;
    PrintModeList curr_mode;

    /* Check each mode step by step */
    curr_mode = all_mode_list;
    while (curr_mode != NULL)
    {
        if (flag[curr_mode->modeid] == 0)
        {
            curr_mode =curr_mode->next;
            continue;
        }

        if (!PrintModeCheck (curr_mode->list, capid, option))
            flag[curr_mode->modeid] = 0;

        j += 1;
        if (j == *passed_mode) /* No need continue checking */
            break;
        curr_mode =curr_mode->next;
    }
    *passed_mode = j;
}

/*****************************************************************************/
CapResultCode
ReferCapRange (CapHandle handle, CapID capid, ElementList *list)
{
    int i = 0, j, k, passed_mode, NUM_OPT = 0;
    int flag[ENTRY_MAX];
    char *option, *option_list[OPTION_MAX], **temp_option_list;
    CapInforPtr cap_infor = (CapInforPtr) handle;
    PrintModeList all_mode_list, curr_mode, temp_mode;
    CapList cap_list;
    ElementList curr_element;
    CapContextList sub_list, context_list = cap_infor->context_list;
    ConsPaperSourceList cons_list = cap_infor->data_list->cons_list;
    sub_list = cap_infor->sub_context_list;

    *list = NULL;
    if (!SearchContextListCap (context_list, capid))
        return NO_ERROR;

    if (capid == P_MEDIA_SIZE)
    {
        char *psid = SearchContextListCap (context_list, P_PAPER_SOURCE);

        ConsPaperSizeList cons_papersize_list;

        while ((cons_list != NULL) && (strcmp (cons_list->psid, psid)))
            cons_list = cons_list->next;
        cons_papersize_list = cons_list->list;
        while (cons_papersize_list != NULL)
        {
            option_list[i++] = cons_papersize_list->ptid;
            cons_papersize_list = cons_papersize_list->next;
        }
        NUM_OPT = i;
    }
    else if (capid == P_WIDE_PRINT) /* Special case */
    {
	    /* FIXME: The wide print is fixed. */
        char *bin_type = SearchContextListCap (context_list, P_PAPER_SOURCE);
        int flag = 0;
        if (!strcmp (bin_type, "AUTO") /* || !strcmp (bin_type, "MANUAL") */ || 
            !strcmp (bin_type, "REAR") || 
            !strcmp (bin_type, "RMANUAL") || 
            !strcmp (bin_type, "MANUALR"))
        {
            char *paper = SearchContextListCap (context_list, P_MEDIA_SIZE);
            if (!strcmp (paper, "ENVC3") || !strcmp (paper, "ENVC4") || !strcmp (paper, "ENVY1")
                    || !strcmp (paper, "ENVY2") || !strcmp (paper, "ENVY3") || !strcmp (paper, "ENVY4"))
                flag = 2;
            else if (!strcmp (paper, "ENV10") || !strcmp (paper, "ENVDL") || !strcmp (paper, "ENVC6") || !strcmp (paper, "ENV5X8"))
                flag = 0;
            else
                flag = 1;
        }

        if ((flag == 0) || (flag == 2))
        {
            if (AddElementListCap (list, "OFF", ""))
                return MEMORY_ERROR;
        }

        if ((flag == 1) || (flag == 2))
        {
            if (AddElementListCap (list, "ON", ""))
                return MEMORY_ERROR;
        }

        return NO_ERROR;
    }
    else if (capid == P_INKSET)
    {
	    cap_list = cap_infor->data_list->cap_list;
	    while ((cap_list != NULL) && (cap_list->capid != capid))
		    cap_list = cap_list->next;
            curr_element = cap_list->list;
	    while (curr_element != NULL) {
		    if (AddElementListCap (list, curr_element->option, curr_element->value)) {
			    return MEMORY_ERROR;
		    }
		    curr_element = curr_element->next;
	    }
	    return NO_ERROR;
    }
    else
    {
        /* Get current print mode */
        if (!strcmp (SearchContextListCap (context_list, P_INK), "COLOR"))
        {
            all_mode_list = cap_infor->data_list->mode_color_list;
        }
        else if (!strcmp (SearchContextListCap (context_list, P_INK), "MONO"))
        {
            all_mode_list = cap_infor->data_list->mode_mono_list;
        }
        else if (!strcmp (SearchContextListCap (context_list, P_INK), "ABWP"))
        {
            all_mode_list = cap_infor->data_list->mode_abwp_list;
        }
        else
        {
            return INVALID_OPTION_ERROR;
        }

        curr_mode = all_mode_list;
        cap_list = cap_infor->data_list->cap_list;

        if (capid == P_PAPER_SOURCE)
        {
            while ((cap_list != NULL) && (cap_list->capid != capid))
                cap_list = cap_list->next;

            curr_element = cap_list->list;
            while (curr_element != NULL)
            {
                curr_mode = all_mode_list;
                while (curr_mode != NULL)
                {
                    if (PrintModeCheck (curr_mode->list, capid, cons_list->psid))
                    {
                        if (AddElementListCap (list, curr_element->option, curr_element->value))
                            return MEMORY_ERROR;
                    }
                    curr_mode =curr_mode->next;
                }
                curr_element = curr_element->next;
            }
            return NO_ERROR;
        }
        else
        {
            passed_mode = cap_infor->modeid;
            while ((curr_mode != NULL) && (curr_mode->modeid != passed_mode))
                curr_mode = curr_mode->next;

            context_list = curr_mode->list; /* The context mode */

            /* Check capids aren't set in the print mode, such as scale, margin, fitopage.. */
            if (!SearchContextListCap (context_list, capid))
            {
                while ((cap_list != NULL) && (cap_list->capid != capid))
                    cap_list = cap_list->next;

                curr_element = cap_list->list;
                while (curr_element != NULL)
                {
                    if (strcmp (curr_element->option, "def"))
                    {
                        if (AddElementListCap (list, curr_element->option, curr_element->value))
                            return MEMORY_ERROR;
                    }
                    curr_element = curr_element->next;
                }

                return NO_ERROR;
            }
            else
            {

                int cur_num_mode = 0;
                curr_mode = all_mode_list;
                while (curr_mode != NULL)
                {
                    flag[++i] = 1;
                    curr_mode = curr_mode->next;
                }
                cur_num_mode = i;
                passed_mode = i;
                context_list = cap_infor->context_list;

                if ((capid == P_MEDIA_TYPE) || (capid == P_RESOLUTION)
		    || (capid == P_QUALITY) || (capid == P_MICRO_WEAVE))
                {
                    /* Checking priority feature firstly  */
                    i = 1;
                    while (i < 7) /* FIXME : Magic number !! */
                    {
                        curr_mode = all_mode_list;
                        option = SearchContextListCap (context_list, priority_table[i]);
                        ReferModeRange (all_mode_list, priority_table[i], option, flag, &cur_num_mode);
                        i += 1;
                        if (priority_table[i] == capid)
                            break;
                    }
                }
                else /* Only get option of this mode */
                {
                    i = 0;
                    while (i < passed_mode)
                        flag[++i] = 0;
                    flag[cap_infor->modeid] = 1;
                }

                /* Get range option for this capid */
                temp_mode = all_mode_list;
                j = k = 0;
                for (i = passed_mode; i > 0; i--)
                    if (flag[i] == 1)
                    {
                        curr_mode = temp_mode;
                        while (curr_mode != NULL)
                        {
                            if (curr_mode->modeid == i)
                            {
                                temp_mode = curr_mode->next;
                                option = SearchContextListCap (curr_mode->list, capid);
                                for (k = 0; k < j; k++)
                                {
                                    if (!strcmp (option_list[k], option))
                                        break;
                                }
                                if (k == j)
                                    option_list[j++] = option;
                                break;
                            }
                            curr_mode = curr_mode->next;
                        }
                    }
                NUM_OPT = j;
            }
        }
    }

    /* Return Elemenlist */
    cap_list = cap_infor->data_list->cap_list;
    while (cap_list != NULL)
    {
        if (cap_list->capid == capid)
        {
            curr_element = cap_list->list;
            while (curr_element != NULL)
            {
                for (i = 0; i < NUM_OPT; i++)
                {
                    temp_option_list = GetAllValueCap (option_list[i]);
                    if (IsExist (curr_element->option, temp_option_list))
                    {
                        if (AddElementListCap (list, curr_element->option, curr_element->value))
                            return MEMORY_ERROR;
                        break;
                    }
                }
                curr_element = curr_element->next;
            }
        }
        cap_list = cap_list->next;
    }

    return NO_ERROR;
}

/*****************************************************************************/
static int
IsExist (char *value, char **mode_value)
{
    while (*mode_value != NULL)
    {
        if (!strcmp (value, *mode_value))
        {
            return 1;
            break;
        }
        mode_value++;
    }

    return 0;
}

/*****************************************************************************/
static int
PrintModeCheck (CapContextList list, int capid, char *option)
{
    CapContextList mode_list = list;
    char **mode_option;

    while ((mode_list != NULL) && (mode_list->capid != capid))
        mode_list = mode_list->next;


    if (mode_list == NULL)
        return 1;

    mode_option = GetAllValueCap (mode_list->option);
    if (IsExist (option, mode_option))
        return 1;

    return 0;
}

/*****************************************************************************/
static int
CombineContextList (CapContextList *first_list, CapContextList second_list)
{
    CapContextList new_list = second_list;
    CapContextList old_list;
    char *new_opt;
    int exist;

    while (new_list != NULL)
    {
        old_list = *first_list;
        exist = 0;

        while (old_list != NULL)
        {
            if (old_list->capid == new_list->capid)
            {
                /* The first position is default value. */
                new_opt = GetValueCap (new_list->option, 1);

                /* FIXME: This is an exception and ugly for inkset. */
                if (new_list->capid != P_INKSET)
                {
                    if (strcmp (old_list->option, new_opt))
                    {
                        free (old_list->option);
                        old_list->option = strdup (new_opt);
                    }
                }
                exist = 1;
                break;
            }
            old_list = old_list->next;
        }
        if (!exist)
            if (AddContextListCap (first_list, new_list->capid, new_list->option))
                return MEMORY_ERROR;
        new_list = new_list->next;
    }
    return NO_ERROR;
}

/*****************************************************************************/
CapResultCode
FixCapContext (CapHandle handle, int fuzzy)
{
    char *option, *ptid, *psid, *temp_opt;
    int flag[ENTRY_MAX];
    int i, j, k, num_opt = 0, passed_mode = 0, last_pm = 0, first_pm = 0;
    CapInforPtr cap_infor = (CapInforPtr) handle;
    CapContextList context_list, sub_context_list, temp_context_list, curr_context;
    PrintModeList curr_mode, all_mode_list;
    ConsPaperSourceList cons_list;
    ConsPaperSizeList cons_papersize_list;
    CapResultCode ret_code = NO_ERROR;

    context_list = cap_infor->context_list;
    sub_context_list = cap_infor->sub_context_list;
    temp_context_list = NULL;

    if (sub_context_list == NULL) /* No need fix cap context */
        return NO_ERROR;

    /* Combine context with sub context list into temp context one */
    while (context_list != NULL)
    {
        AddContextListCap (&temp_context_list, context_list->capid, context_list->option);
        context_list = context_list->next;
    }

    /* Set Inkset */
    option = SearchContextListCap (sub_context_list, P_INKSET);
    if (option)
    {
        if (AddContextListCap (&temp_context_list, P_INKSET, option))
            return MEMORY_ERROR;
    }

    if (CombineContextList (&temp_context_list, sub_context_list))
        return MEMORY_ERROR;

    context_list = cap_infor->context_list;         /* Re-get the context list */


    /* Check Ink mode firstly */
    for (i = 0; i < ENTRY_MAX; i++)
        flag[i] = 0;

    if (!strcmp (SearchContextListCap (temp_context_list, P_INK), "COLOR"))
    {
        all_mode_list = cap_infor->data_list->mode_color_list;
    }
    else if (!strcmp (SearchContextListCap (temp_context_list, P_INK), "MONO"))
    {
        all_mode_list = cap_infor->data_list->mode_mono_list;
    }
    else if (!strcmp (SearchContextListCap (temp_context_list, P_INK), "ABWP"))
    {
        all_mode_list = cap_infor->data_list->mode_abwp_list;
    }
    else
    {
        return INVALID_OPTION_ERROR;
    }


    ptid = SearchContextListCap (temp_context_list, P_MEDIA_SIZE);
    psid = SearchContextListCap (temp_context_list, P_PAPER_SOURCE);
    cons_list = cap_infor->data_list->cons_list;

    if (ptid == 0 || psid == 0)
    {
        return INVALID_OPTION_ERROR;
    }

    /* Check PaperSize, WidePrint */
    while ((cons_list != NULL) && strcmp (cons_list->psid, psid))
    {
        cons_list = cons_list->next;
    }

    cons_papersize_list = cons_list->list;

    while ((cons_papersize_list != NULL) && strcmp (cons_papersize_list->ptid, ptid))
    {
        if (cons_papersize_list->next == NULL)
            ptid = cons_papersize_list->ptid; /* Get the first option in list */
        cons_papersize_list = cons_papersize_list->next;
    }

    if (!cons_papersize_list)
    {
        if (!fuzzy)
            return COMBINE_OPTION_ERROR;
        else
            AddContextListCap (&temp_context_list, P_MEDIA_SIZE, ptid);
    }

    if (!strcmp (psid, "AUTO") /* || !strcmp (psid, "MANUAL") */ || 
        !strcmp (psid, "REAR") || 
        !strcmp (psid, "RMANUAL") ||
        !strcmp (psid, "MANUALR"))
    {
        if (!strcmp (ptid, "ENVC3") || !strcmp (ptid, "ENVC4") || !strcmp (ptid, "ENVY1")
                || !strcmp (ptid, "ENVY2") || !strcmp (ptid, "ENVY3") || !strcmp (ptid, "ENVY4"))
            option = NULL;
        else if (!strcmp (ptid, "ENV10") || !strcmp (ptid, "ENVDL") || !strcmp (ptid, "ENVC6")
                 || !strcmp (ptid, "ENV5X8"))
            option = "OFF";
        else
            option = "ON";
    }
    else
        option = "OFF";

    temp_opt = SearchContextListCap (temp_context_list, P_WIDE_PRINT);
    if (option)
        if (option != NULL && strcmp (option, temp_opt))
        {
            if (cap_infor->filter_type == 1)
            {
                if (!fuzzy)
                    return COMBINE_OPTION_ERROR;
                else
                    AddContextListCap (&temp_context_list, P_WIDE_PRINT, option);
            }
            else /* filter_type == 2 */
            {
                if (!strcmp (temp_opt, "ON"))
                    return COMBINE_OPTION_ERROR;
                else
                    AddContextListCap (&temp_context_list, P_WIDE_PRINT, option);
            }
        }

    curr_context = temp_context_list;
    curr_mode = all_mode_list;

    if (cap_infor->filter_type == 1) /* png filter */
    {
        /* Only check combine options for all options of print mode */
        /* Checking follow priority of features */
        num_opt = 7; /* Priority 6 options - FIXME: Magic number !! */

        curr_context = all_mode_list->list;
        while (curr_context != NULL)
        {
            if ((curr_context->capid != P_INK)
		&& (curr_context->capid != P_PAPER_SOURCE)
		&& (curr_context->capid != P_MEDIA_TYPE)
		&& (curr_context->capid != P_RESOLUTION)
		&& (curr_context->capid != P_INKSET)
		&& (curr_context->capid != P_QUALITY)
		&& (curr_context->capid != P_MICRO_WEAVE))
		    priority_table[num_opt++] = curr_context->capid;

            curr_context = curr_context->next;
        };
        /* Get the number of print mode */
        i = 1; /* Starting the 1st */
        while (curr_mode != NULL)
        {
            flag[i++]
            = 1;
            curr_mode = curr_mode->next;
        }

        passed_mode = i - 1; /* The number of mode is passed */

        for (i = 1; i < num_opt; i++) /* FIXME: This doesn't explain that excludes P_INK. */
        {
            option = SearchContextListCap (temp_context_list, priority_table[i]);

            /* Check each mode step by step */
            curr_mode = all_mode_list;
            j = k = first_pm = 0;
            while (curr_mode != NULL)
            {
                if (flag[curr_mode->modeid] == 0)
                {
                    curr_mode =curr_mode->next;
                    continue;
                }

                if (!PrintModeCheck (curr_mode->list, priority_table[i], option))
                    flag[curr_mode->modeid] = 0;
                else
                {
                    k += 1;
                    first_pm = curr_mode->modeid;
                }
                j += 1;

                if (j == passed_mode) /* No need continue checking */
                    break;
                curr_mode =curr_mode->next;
            }
            passed_mode = k;
            if (first_pm > 0)
                last_pm = first_pm;
            if (passed_mode == 0) /* With option, no mode is passed! */
                break;
        }
    }
    else /* Raw filter */
    {
        option = SearchContextListCap (curr_context, P_QUALITY_RAW);
        while ((curr_mode != NULL) && (strcmp (curr_mode->mode_name, option)))
            curr_mode = curr_mode->next;
        /* Check mode */
        if (curr_mode == NULL)
            return COMBINE_OPTION_ERROR;
        /* Check paper source */
        strcpy (option, SearchContextListCap (curr_context, P_PAPER_SOURCE));
        if (!PrintModeCheck (curr_mode->list, P_PAPER_SOURCE, option))
            return COMBINE_OPTION_ERROR;

        if (CombineContextList (&temp_context_list, curr_mode->list))
            return MEMORY_ERROR;
        /* Previous cmd made paper src change, should be update paper src */
        AddContextListCap (&temp_context_list, P_PAPER_SOURCE, option);

        DeleteContextListCap (&cap_infor->context_list);
        DeleteContextListCap (&cap_infor->sub_context_list);
        cap_infor->context_list = temp_context_list;

        return NO_ERROR;
    }

    /* Illegal combine option*/
    if (passed_mode == 0)
    {
        if (!fuzzy) /* Call from pips-filter */
        {
            DeleteContextListCap (&cap_infor->sub_context_list);
            ret_code = COMBINE_OPTION_ERROR;
        }
        else /* Call from pips-tool */
        { /* Find the nearest legal by fuzzy way... */
            if (last_pm == 0) /* No passed options! */
                ret_code = COMBINE_OPTION_ERROR; /* Keep old context */
            else
            {
                cap_infor->modeid = last_pm;
                /* Update the first passed mode to temp context list */
                curr_mode = all_mode_list;
                while ((curr_mode != NULL) && (curr_mode->modeid != last_pm))
                    curr_mode = curr_mode->next;
                /* Get default value */
                option = strdup (psid);
                CombineContextList (&temp_context_list, curr_mode->list);

                AddContextListCap (&temp_context_list, P_PAPER_SOURCE, option);
                DeleteContextListCap (&cap_infor->context_list);
                cap_infor->context_list = temp_context_list;
		if (option != NULL)
		{
		    free (option);
		    option = NULL;
		}
            }
        }
    }
    else /* Combine OK */
    {
        cap_infor->modeid = last_pm;
        DeleteContextListCap (&cap_infor->context_list);
        cap_infor->context_list = temp_context_list;
    }
    DeleteContextListCap (&cap_infor->sub_context_list);

    return ret_code;
}

/*****************************************************************************/
CapResultCode
ReferCapContext (CapHandle handle, CapContextList *list)
{
    CapInforPtr cap_infor = (CapInforPtr) handle;
    CapContextList context_list;

    /* Current other_list that included name_def_list and file lists */
    context_list = cap_infor->data_list->other_list;
    while (context_list != NULL)
    {
        if (AddContextListCap (list, context_list->capid, context_list->option))
            return MEMORY_ERROR;
        context_list = context_list->next;
    }

    /* Current context_list */
    context_list = cap_infor->context_list;
    while (context_list != NULL)
    {
        if (AddContextListCap (list, context_list->capid, context_list->option))
            return MEMORY_ERROR;
        context_list = context_list->next;
    }
    return NO_ERROR;
}

/*****************************************************************************/
CapResultCode
GetPaperRsc (CapHandle handle, char *psid, char *ptid, LP_PS_STRUCT paper)
{
    CapInforPtr cap_infor = (CapInforPtr) handle;
    PaperSizeList paper_list;
    ConsPaperSourceList cons_list;
    ConsPaperSizeList cons_paper_list;
    CapList common_list;
    int i = 0;
    char *size = NULL;

    cons_list = cap_infor->data_list->cons_list;
    while ((cons_list != NULL) && (strcmp (cons_list->psid, psid)))
        cons_list = cons_list->next;

    cons_paper_list = cons_list->list;
    while (cons_paper_list != NULL)
    {
        if (!strcmp (cons_paper_list->ptid, ptid))
        {
            paper->area.x = cons_paper_list->pa_width;
            paper->area.y = cons_paper_list->pa_length;
            paper->def_area.x = cons_paper_list->pa_width;
            paper->def_area.y = cons_paper_list->pa_length;
            paper->margin.x = cons_paper_list->left_margin;
            paper->margin.y = cons_paper_list->top_margin;
            i = 1;
            break;
        }
        cons_paper_list = cons_paper_list->next;
    }

    if (i == 0)
        return INVALID_OPTION_ERROR;

    paper_list = cap_infor->data_list->papersize_list;
    while (paper_list != NULL)
    {
        if (!strcmp (paper_list->ptid, ptid))
        {
            paper->size.x = paper_list->width;
            paper->size.y = paper_list->height;
            break;
        }
        paper_list = paper_list->next;
    }

    common_list = cap_infor->data_list->common_list;
    i = 0;
    while (common_list != NULL)
    {
        if (common_list->capid == P_MARGIN_X)
        {
            i += 1;
            if ((size = SearchElementListCap (common_list->list, "def")))
                paper->def_margin.x = atol (size);
        }
        else if (common_list->capid == P_MARGIN_Y)
        {
            i += 1;
            if ((size = SearchElementListCap (common_list->list, "def")))
                paper->def_margin.y = atol (size);
        }
        if (i == 2)
            break;

        common_list = common_list->next;
    }

    return NO_ERROR;
}

/*****************************************************************************/
CapResultCode
LoadCapContext (CapHandle handle, char *filename)
{
    CapInforPtr cap_infor = (CapInforPtr) handle;
    CapResultCode ret_code = 0;
    CapContextList context_list = cap_infor->context_list;
    char *printer_name = SearchContextListCap (cap_infor->data_list->name_def_list, P_MODEL_NAME);

    ret_code = LoadPipsrcFile (printer_name, filename, &context_list);
    return ret_code;
}

/*****************************************************************************/
CapResultCode
SaveCapContext (CapHandle handle, char *filename, int save_type)
{
    int ret_code = 0;
    CapInforPtr cap_infor = (CapInforPtr) handle;
    char *printer_name = SearchContextListCap (cap_infor->data_list->name_def_list, P_MODEL_NAME);

    if (cap_infor->sub_context_list)
        CombineContextList (&cap_infor->context_list, cap_infor->sub_context_list);

    ret_code = SavePipsrcFile (printer_name, filename, cap_infor->context_list, save_type);

    return ret_code;
}
/*****************************************************************************/
