/*
 * Photo Image Print System
 * Copyright (C) 2002-2008 AVASYS Corporation.
 * Copyright (C) Seiko Eosib Corporation 2002-2009.
 *
 *  This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * As a special exception, EPSON AVASYS Corporation gives permission to
 * link the code of this program with libraries which are covered by
 * the EPSON AVASYS Public License and distribute their linked
 * combinations.  You must obey the GNU General Public License in all
 * respects for all of the code used other than the libraries which
 * are covered by EPSON AVASYS Public License.
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "pips-prtlib.h"
#include "pips-caplib.h"
#include "pips-filter.h"
#include "pfimg.h"
#include "setup.h"
#include "pipsmem.h"
#include "pipserr.h"
#include "rawfilter.h"
#include "debug.h"

#define MAX_ARGC 15
#define MIN_ARGC 13

static PairStr replace_media[] = {
  { "A4",           "A4_AUTO" },
  { "Letter",       "LT_AUTO" },
  { "A6",           "A6_AUTO" },
  { "A5",           "A5_AUTO" },
  { "B5",           "B5_AUTO" },
  { "Env10",        "ENV10_AUTO" },
  { "EnvDL",        "ENVDL_AUTO" },
  { "EnvC6",        "ENVC6_AUTO" },
  { "Legal",        "LGL_AUTO" },
  { "Postcard",     "PC_AUTO" },
  { NULL, NULL }
};

static int raster_to_printer (char*, CapContextList, int, int);
static int raw_change_prtinfo (HANDLE, HANDLE, CapContextList*);

/* Usage: pips-filter -f raw -p <printer_name> -in <ink_mode> -ms <Paper_Size> -ql <Quality_Mode> [<Rawfile>]*/
int
RawFilter (int argc, char **argv)
{
    char cap_path[] = LIB_CAP_PATH;
    char *model_name, *alias_name, *libprt_path, *ink_set = NULL;
    char err_data[MAX_SIZE];
    char *newargv[MAX_ARGC];
    int i, ret_code = 0;
    int cupsWidth, cupsHeight;
    /* 2004.04.15 added for 'cancel page' */

    CapContextList p_caplist = NULL;
    HANDLE lib_handle = NULL;
    HANDLE cap_handle = NULL;
    OPENCAP_FUNC OpenCap;
    SETCAPLIST_FUNC SetCapList;
    FIXCAPCTX_FUNC FixCap;
    REFCAPCTX_FUNC ReferCap;

    /* attach point */
#ifdef DEBUG
    int flag = 1;
    while (flag) { sleep(3); }
    DEBUG_START;
#endif /* DEBUG */

    /* Initialize */
    model_name =  libprt_path = alias_name = NULL;
    if ((argc < MIN_ARGC || argc > MAX_ARGC))
    {
        fprintf (stderr, "Usage: pips-filter -r raw -p <printer_name> -in <ink_mode> \
                 -ms <Paper_Size> -ql <Quality_Mode> <Header.cupsWidth> <Header.cupsHeight> -is <ink_set>\n");
        return 1;
    }

    if ((i = GetGlobalOpt (argc, argv, "p")))
    {
        if (i == -1)
            return 1;
        model_name = argv[i];
    }

    if (model_name == NULL)
    {
        if (!(model_name = GetDefaultPrinter(1)))
        {
            FILTER_ERROR_HANDLER (DEF_PRINTER_NOT_FOUND, NULL);
            char **model_list = NULL;
            model_list = GetModelListName (PNG_TYPE);
            if (*model_list)
            {
                model_name = *model_list;
            }

            else
            {
                FILTER_ERROR_HANDLER (PRINTER_NOT_FOUND, NULL);
                return 1;
            }
        }
    }

    alias_name = GetAliasName(model_name);
    if (alias_name == NULL) {
	FILTER_ERROR_HANDLER (PRINTER_NOT_FOUND, NULL);
	return 1;
    }

    if (!(libprt_path = GetLibPrtPath (model_name)))
        return 1;

    lib_handle = lt_dlopen (cap_path);
    if (lib_handle == NULL)
    {
        FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror());
        return 1;
    }

    memset (err_data, '\0', sizeof (err_data));
    OpenCap = (OPENCAP_FUNC) lt_dlsym (lib_handle, "OpenCap");
    if (OpenCap == NULL)
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror());
        return 1;
    }
    if ((ret_code = (*OpenCap) (&cap_handle, model_name, RAW_TYPE, err_data)))
    {
        if (ret_code != PIPSRC_NOT_FOUND)
        {
            lt_dlclose (lib_handle);
            FILTER_ERROR_HANDLER (ret_code, err_data);
            return 1;
        }
        ret_code = 0;
    }

    cupsWidth = atoi (argv [11]);
    cupsHeight = atoi (argv[12]);
    if (argc == MAX_ARGC)
        ink_set = argv[14];
    argc = 11;

/* failsafe check */
#if 11 >= MAX_ARGC
#error "MAX_ARGC ERROR"
#endif

    /* Replace some media's parameter */
    {
      int j, k;
      for(j = 0; j < argc; j++)
      {
        newargv[j] = NULL;
        for(k = 0; ( NULL != replace_media[k].old ) && ( 0 != strcmp(argv[j], replace_media[k].old) ) ; k++)
        {
          /* empty loop */
        }
        if(NULL != replace_media[k].old)
        {
          newargv[j] = strdup(replace_media[k].new);
        }
        else
        {
          newargv[j] = strdup(argv[j]);
        }
        if(NULL == newargv[j])
        {
          FILTER_ERROR_HANDLER (MEMORY_ERROR, argv[j]);
          return 1;
        }
      }
      newargv[argc] = NULL;
    }

    /* Set option list for SubContext */
    SetCapList = (SETCAPLIST_FUNC)lt_dlsym (lib_handle, "SetCapList");
    if (SetCapList == NULL )
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror());
        return 1;
    }

    ret_code = (*SetCapList) (cap_handle, newargv, argc, err_data); /* Update ContextList */
    {
      int j;
      for(j = 0; j < argc; j++)
      {
        if(NULL != newargv[j])
        {
          free(newargv[j]);
          newargv[j] = NULL;
        }
      }
    }
    if (ret_code)
    {
      lt_dlclose (lib_handle);
      FILTER_ERROR_HANDLER (ret_code, err_data);
      return 1;
    }

    /* FixCapContext for getting the correct CapContext */
    FixCap = (FIXCAPCTX_FUNC)lt_dlsym (lib_handle, "FixCapContext");
    if (FixCap == NULL )
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror());
        return 1;
    }
    if ((ret_code = (*FixCap) (cap_handle, FALSE)))
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (ret_code, cap_path);
        return 1;
    }

    /* Get CapContextList */
    ReferCap = (REFCAPCTX_FUNC)lt_dlsym (lib_handle, "ReferCapContext");
    if (ReferCap == NULL )
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror());
        return 1;
    }
    if ((ret_code = (*ReferCap) (cap_handle, &p_caplist)))
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (ret_code, cap_path);
        return 1;
    }
    /*if (!(input_file = GetCapValue (p_caplist, P_INPUT_FILE)))
    {
        input_file = (char *)malloc( strlen(STD_IN_NAME) + 1);
        strcpy (input_file, STD_IN_NAME);
    }
    else
    {	// Check input data file /
        if (CheckExistFile (input_file))
        {
            FILTER_ERROR_HANDLER (FILE_NOT_FOUND_ERROR, input_file);
            return 1;
        }
    } */
    if (ink_set)
        if (UpdateContextList (p_caplist, P_INKSET, ink_set))
        {
            FILTER_ERROR_HANDLER (ret_code, cap_path);
            return 1;
        }

    if (raw_change_prtinfo (lib_handle, cap_handle, &p_caplist))
        return 1;

    if (raster_to_printer (libprt_path, p_caplist, cupsWidth, cupsHeight))
        return 1;

    /* close library */
    if (lib_handle)
    {
        if (lt_dlclose (lib_handle) == -1)
        {
            FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror ());
            return 1;
        }
    }

    return NO_ERROR;
}

static int
raw_change_prtinfo (HANDLE lib_handle, HANDLE cap_handle, CapContextList *p_caplist)
{
    char cap_path[] = LIB_CAP_PATH;
    int resol;
    GETPAPERRSC_FUNC GetPaperRsc;
    LP_PS_STRUCT paper_conf = NULL;

    paper_conf = pips_malloc (PS_STRUCT, 1);

    /* Get Paper Size correspondence with Paper Source */
    GetPaperRsc = (GETPAPERRSC_FUNC)lt_dlsym (lib_handle, "GetPaperRsc");
    if (GetPaperRsc == NULL)
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror());
        return 1;
    }

    if ((*GetPaperRsc) (cap_handle, GetCapValue (*p_caplist, P_PAPER_SOURCE), GetCapValue (*p_caplist, P_MEDIA_SIZE), paper_conf))
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, cap_path);
        return 1;
    }

    if (!(resol = atoi(GetCapValue (*p_caplist, P_RESOLUTION))))
        return 1;

    paper_conf->size.x   = paper_conf->size.x * resol / 360;
    paper_conf->size.y   = paper_conf->size.y * resol / 360;
    paper_conf->area.x   = paper_conf->area.x * resol / 360;
    paper_conf->area.y = paper_conf->area.y * resol / 360;
    paper_conf->margin.x = paper_conf->margin.x * resol / 360;
    paper_conf->margin.y = paper_conf->margin.y * resol / 360;

    if (AddPaperRsc (p_caplist, paper_conf, FALSE))
        return 1;

    if (AddContextList (p_caplist, P_BRIGHTNESS, "0") ||
        AddContextList (p_caplist, P_CONTRAST, "0") ||
        AddContextList (p_caplist, P_SATURATION, "0") ||
        AddContextList (p_caplist, P_STRENGTH_CYAN, "0") ||
        AddContextList (p_caplist, P_STRENGTH_MAGENTA, "0") ||
        AddContextList (p_caplist, P_STRENGTH_YELLOW, "0"))
    {
        FILTER_ERROR_HANDLER (MEMORY_ERROR, NULL);
        return 1;
    }

    return NO_ERROR;
}

static int
raster_to_printer (char *libprt_path, CapContextList p_caplist, int image_width, int image_height)
{
    int i, j, cancel = 0;
    int byte_per_pixel, band_line, conv_status;
    long area_x, area_y, bytes_per_line;
    double margin_x, margin_y;
    long read_size = 0;
    char *image_raw, *wide_print;
    unsigned char *band;
    FILE *output_fp = stdout;
    PrtBand band_str;
    CapHandle drv_handle;
    HANDLE lib_handle;
    OPEN_FUNC OpenDriver;
    STARTJOB_FUNC StartJob;
    STARTPAGE_FUNC StartPage;
    BANDOUT_FUNC BandOut;
    ENDPAGE_FUNC EndPage;
    ENDJOB_FUNC EndJob;
    CLOSE_FUNC CloseDriver;

    /* Close written pipe direction */
#ifdef DEBUG
    DEBUG_INIT_JOB (p_caplist);
#endif /* DEBUG */

    lib_handle = lt_dlopen (libprt_path);
    if (lib_handle == NULL)
    {
        FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror());
        return 1;
    }
    /* Setting of library function */
    OpenDriver  = (OPEN_FUNC)lt_dlsym (lib_handle, "OpenDriver");
    StartJob    = (STARTJOB_FUNC)lt_dlsym (lib_handle, "StartJob");
    StartPage   = (STARTPAGE_FUNC)lt_dlsym (lib_handle, "StartPage");
    BandOut     = (BANDOUT_FUNC)lt_dlsym (lib_handle, "BandOutput");
    EndPage     = (ENDPAGE_FUNC)lt_dlsym (lib_handle, "EndPage");
    EndJob      = (ENDJOB_FUNC)lt_dlsym (lib_handle, "EndJob");
    CloseDriver = (CLOSE_FUNC)lt_dlsym (lib_handle, "CloseDriver");

    if (!OpenDriver
            || !StartJob
            || !StartPage
            || !BandOut
            || !EndPage
            || !EndJob
            || !CloseDriver)
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror());
    }

    /* Open library */
    if ((*OpenDriver)(&drv_handle))
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, libprt_path);
        return 1;
    }
    conv_status = CONV_STATUS_OPEN;
    /* Init Job */
    if ((*StartJob)(drv_handle, p_caplist, output_fp))
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, libprt_path);
        return 1;
    }
    conv_status = CONV_STATUS_JOB_START;
    band_line = 2;
    if (strcmp (GetCapValue (p_caplist, P_INK), "COLOR") == 0)
        byte_per_pixel = 3;
    else
        byte_per_pixel = 1;

    /* Setup band struct */
    memset (&band_str, 0, sizeof (PrtBand));
    area_x = atol (GetCapValue (p_caplist, P_PAPER_AREA_X));
    area_y = atol (GetCapValue (p_caplist, P_PAPER_AREA_Y));
    if ((wide_print = GetCapValue (p_caplist, P_WIDE_PRINT))) /* Support wide print */
    {
        if (!strcmp (wide_print, "ON"))
            area_y = atol (GetCapValue (p_caplist, P_PAPER_SIZE_Y)) - atol (GetCapValue (p_caplist, P_MARGIN_Y))
                        - 42 * atol (GetCapValue (p_caplist, P_RESOLUTION)) / 360;
    }

    band_str.clip.width = area_x;
    band_str.clip.height = band_line;
    band_str.rows = (Lp_RasterRows) pips_malloc (RasterRows, 1);
    band_str.rows->size = WIDTH_BYTES (area_x * byte_per_pixel * 8);
    band_str.at_rows = NULL;

    band = pips_malloc (char, band_str.rows->size * band_line);
    memset (band, 0xff, band_str.rows->size * band_line);
    band_str.rows->data = band;
    margin_x = (double) area_x / image_width;
    margin_y = (double) area_y / image_height;
    bytes_per_line = WIDTH_BYTES(image_width * byte_per_pixel * 8);
    image_raw = pips_calloc (char, bytes_per_line);

#ifdef DEBUG
    DEBUG_INIT_PAGE (p_caplist);
    DEBUG_BAND_OUT (band_str);
#endif /* DEBUG */

    while ((read_size = read (STDIN_FILENO, image_raw, bytes_per_line)) > 0)
    {
        long x_count, y_count;
        int band_line_count;
        y_count = 0;
        band_line_count = 0;
        band_str.clip.height = band_line;
        /* InitPage */
        if ((*StartPage) (drv_handle, p_caplist))
        {
            lt_dlclose (lib_handle);
            FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, libprt_path);
            return 1;
        }
        conv_status = CONV_STATUS_PAGE_START;
        for (i = 0; i < area_y; i ++)
        {
            char *line;
            line = band + (band_str.rows->size * band_line_count);
            while ((0 == y_count) || ((i > (margin_y * y_count) - 1) && (y_count < image_height)))
            {
                while (read_size < bytes_per_line)
                {
                    size_t rsize;
                    rsize = read (STDIN_FILENO, image_raw + read_size, bytes_per_line - read_size);
                    if (rsize < 0)
                    {
                        if ((errno != EINTR) && (errno != EAGAIN) && (errno != EIO))
                        {
                            /* 2004.04.15		*/
                            /* error then quit	*/
                            /* don't care err = -1  */
                            cancel = 1;
                            goto quit;
                            /* not reached		*/
                            break;
                        }
                        usleep(50000);
                    }
                    else
                        read_size += rsize;
                }

                /* 2004.04.01                                              */
                /* for "skip reading"                                      */
                /* modified y_count count up condition and clear read_size */
                if (read_size == bytes_per_line)
                {
                    y_count++;
                    /* need to clear read_size               */
                    /* not clear, data still remains on pipe */
                    read_size = 0;
                }
            }
            read_size = 0;

            if (margin_x == 1)
                memcpy (line, image_raw, bytes_per_line);
            else
            {
                x_count = 0;

                for (j = 0; j < area_x; j ++)
                {
                    while (x_count == 0 || j > margin_x * x_count)
                        x_count ++;

                    memcpy (line + (j * byte_per_pixel),image_raw + ((x_count - 1) * byte_per_pixel), byte_per_pixel);
                }
            }

            band_line_count ++;

            if (band_line_count >= band_line)
            {
                if (BandOut (drv_handle, band_str.clip, band_str.rows, NULL))
                {
                    lt_dlclose (lib_handle);
                    FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, libprt_path);
                    return 1;
                }
                band_line_count = 0;
            }
        }
        if (band_line_count > 0)
        {
            band_str.clip.height = band_line_count;
            if (BandOut (drv_handle, band_str.clip, band_str.rows, NULL))
            {
                lt_dlclose (lib_handle);
                FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, libprt_path);
                return 1;
            }
        }
        conv_status = CONV_STATUS_BAND_OUT;

        /* TerminalPage */
        if ((*EndPage) (drv_handle, 0))
        {
            lt_dlclose (lib_handle);
            FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, libprt_path);
            return 1;
        }
        conv_status = CONV_STATUS_PAGE_END;
    }

#ifdef DEBUG
    DEDBUG_END;
#endif /* DEBUG */

    /* 2004.04.15 for 'error' */
quit:
    ;
    if (cancel)
    {
        /* TerminalPage */
        if ((*EndPage) (drv_handle, cancel))
        {
            lt_dlclose (lib_handle);
            FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, libprt_path);
            return 1;
        }

    }
    /* Destroy Job */
    if ((*EndJob)(drv_handle))
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, libprt_path);
        return 1;
    }
    conv_status = CONV_STATUS_JOB_END;
    /* Close Driver */
    if ((*CloseDriver) (drv_handle))
    {
        lt_dlclose (lib_handle);
        FILTER_ERROR_HANDLER (LIB_INTERNAL_ERROR, libprt_path);
        return 1;
    }
    conv_status = CONV_STATUS_CLOSE;
    if (lib_handle)
    {
        if (lt_dlclose (lib_handle) == -1)
        {
            FILTER_ERROR_HANDLER (LIB_ERROR, (char*)lt_dlerror ());
            return 1;
        }
    }

    return 0;
}

